// a simple interface for turtle field
// Yoshinori Hayakawa
// 10-JUL-2013 version 
// 30-JUN-2017

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.Socket;


public class Turtle {

    public class TeamMessage {
	String message ;
	int time ;
    }

    private Socket socket;
    private int port ;
    private BufferedReader sock_in ;
    private PrintWriter sock_out ;
    private String last_message ;
    private int team ;

    private static final int EVENT_HIT_BY_BULLET = (0x00000001) ;
    private static final int EVENT_RUN_INTO_TURTLE=(0x00000002) ;
    private static final int EVENT_RUN_INTO_DONUT= (0x00000004) ;
    private static final int EVENT_RUN_INTO_STONE= (0x00000008) ;
    private static final int EVENT_RUN_INTO_WALL=  (0x00000010) ;
    private static final int EVENT_FOUND_COIN=     (0x00000020) ;
    private static final int EVENT_DETECTED_BY_FINDER= (0x00000040) ;
    private static final int EVENT_DETECTED_BY_RADAR=  (0x00000080) ;
    private static final int EVENT_GOT_MESSAGE=        (0x00000100) ;

    Turtle(String hostname) {
	port = 2222 ;
	try {
            socket = new Socket(hostname, port);
	    sock_in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
	    sock_out = new PrintWriter(socket.getOutputStream(), true);
	} catch (IOException e) {
	    System.err.println("cannot connect to " + hostname) ;
	    System.err.println("seems turtle field is not ready...") ;
	    System.exit(0) ; 
	} 
    }

    private int sendMessageAndReceiveAck(String cmd) {
	int eventcode,timecode,rc=0 ;
	try {
	    sock_out.println(cmd);
	    last_message = sock_in.readLine();
	} catch (IOException e) {
	    closeSocket() ;
	    System.err.println("I was killed.. bye!!") ;
	    System.exit(0) ; 
	}
	try {
	    String[] ack = last_message.split(" ",0) ;
	    eventcode = Integer.valueOf(ack[ack.length-1]) ;
	    timecode = Integer.valueOf(ack[ack.length-2]) ;
	    if ( ack[ack.length-4].equals("OK")) rc = 0 ; else rc = -1 ;
	    if (eventcode!=0) {
		handleEvent(eventcode,timecode) ;
	    }
	} catch (NullPointerException e) {
	    System.err.println("I was killed... bye!!") ;
	    System.exit(0) ; 
	}
	return rc ;
    }

    int enable_event(int eventcode) {
	return sendMessageAndReceiveAck("Y + " + eventcode) ;
    }

    int disable_event(int eventcode) {
	return sendMessageAndReceiveAck("Y - " + eventcode) ;
    }


    private void handleEvent(int eventcode, int timecode) {
	String msg_stack = new String(last_message) ;
	int curr_team = team ;
	if ((eventcode & EVENT_HIT_BY_BULLET) !=0) {
	    disable_event(EVENT_HIT_BY_BULLET) ;
	    hit_by_bullet(timecode) ;
	    enable_event(EVENT_HIT_BY_BULLET) ;
	}
	if ((eventcode & EVENT_RUN_INTO_TURTLE) !=0) {
	    disable_event(EVENT_RUN_INTO_TURTLE) ;
	    run_into_turtle(timecode) ;
	    enable_event(EVENT_RUN_INTO_TURTLE) ;
	}
	if ((eventcode & EVENT_RUN_INTO_DONUT) !=0) {
	    disable_event(EVENT_RUN_INTO_DONUT) ;
	    run_into_donut(timecode) ;
	    enable_event(EVENT_RUN_INTO_DONUT) ;
	}
	if ((eventcode & EVENT_RUN_INTO_STONE) !=0) {
	    disable_event(EVENT_RUN_INTO_STONE) ;
	    run_into_stone(timecode) ;
	    enable_event(EVENT_RUN_INTO_STONE) ;
	}
	if ((eventcode & EVENT_RUN_INTO_WALL) !=0) {
	    disable_event(EVENT_RUN_INTO_WALL) ;
	    run_into_wall(timecode) ;
	    enable_event(EVENT_RUN_INTO_WALL) ;
	}
	if ((eventcode & EVENT_FOUND_COIN) !=0) {
	    disable_event(EVENT_FOUND_COIN) ;
	    found_coin(timecode) ;
	    enable_event(EVENT_FOUND_COIN) ;
	}
	if ((eventcode & EVENT_DETECTED_BY_FINDER) !=0) {
	    disable_event(EVENT_DETECTED_BY_FINDER) ;
	    detected_by_finder(timecode) ;
	    enable_event(EVENT_DETECTED_BY_FINDER) ;
	}
	if ((eventcode & EVENT_DETECTED_BY_RADAR) !=0) {
	    disable_event(EVENT_DETECTED_BY_RADAR) ;
	    detected_by_radar(timecode) ;
	    enable_event(EVENT_DETECTED_BY_RADAR) ;
	}
	if ((eventcode & EVENT_GOT_MESSAGE) !=0) {
	    disable_event(EVENT_GOT_MESSAGE) ;
	    got_message(timecode) ;
	    enable_event(EVENT_GOT_MESSAGE) ;
	}
	last_message = msg_stack ;
	team = curr_team ;
    }

    public void hit_by_bullet(int timecode) {}
    public void run_into_turtle(int timecode) {}
    public void run_into_donut(int timecode) {}
    public void run_into_doughnut(int timecode) {run_into_donut(timecode);}
    public void run_into_stone(int timecode) {}
    public void run_into_wall(int timecode) {}
    public void found_coin(int timecode) {}
    public void detected_by_finder(int timecode) {}
    public void detected_by_radar(int timecode) {}
    public void got_message(int timecode) {}

    private void closeSocket() {
	if (socket != null && !socket.isClosed()) {
	    try {
		socket.close();
	    } catch (IOException e) {
		e.printStackTrace();
	    }
	}
    }

    private String convert_str2hex(String str) {
	StringBuilder sb = new StringBuilder();
	byte[] bytes = str.getBytes() ;
	for (byte b : bytes) {
	    sb.append(Integer.toHexString((b & 0xF0) >> 4));
	    sb.append(Integer.toHexString(b & 0xF));
	}
	return sb.toString() ;
    }

    private int byte2int(byte c)
    {
	int r=0 ;
	if (c >= '0' && c<= '9') r = c - '0' ;
	else if (c>='a' && c<='f') r = c - 'a' + 10 ;
	else if (c>='A' && c<='F') r = c - 'A' + 10 ;
	return r ;
    }
    
    private String convert_hex2str(String hex) {
	int cnt ; int d0=0,d1=0 ; char c ;
	StringBuilder sb = new StringBuilder();
	byte[] bytes = hex.getBytes() ;
	cnt=0 ;
	for (byte b : bytes) {
	    if (b==' ') continue ;
	    if (cnt%2==0) d1 = byte2int(b) ;
	    else {
		d0 = byte2int(b) ;
		c = (char) (d1*16 + d0) ;
		sb.append(c) ;
	    }
	    cnt++ ;
	}
	return new String(sb) ;
    }


    public int home() {
	return sendMessageAndReceiveAck("C H") ;
    }

    public int rst() {
	return sendMessageAndReceiveAck("C") ;
    }

    public int cf() {
	return sendMessageAndReceiveAck("C F") ;
    }

    public int clr() {
	return sendMessageAndReceiveAck("C A") ;
    }	

    public int fd(double step) {
	return sendMessageAndReceiveAck("F " + step) ;
    }

    public int bk(double step) {
	return sendMessageAndReceiveAck("B " + step) ;
    }

    public int jump(double x, double y) {
	return sendMessageAndReceiveAck("J " + x + " " + y) ;
    }

    public int rjump(double dx, double dy) {
	return sendMessageAndReceiveAck("G " + dx + " " + dy) ;
    }

    public int rt(double angle) {
	return sendMessageAndReceiveAck("T R " + angle) ;
    }

    public int lt(double angle) {
	return sendMessageAndReceiveAck("T L " + angle) ;
    }

    public int east() {
	return sendMessageAndReceiveAck("T E") ;
    }

    public int west() {
	return sendMessageAndReceiveAck("T W") ;
    }

    public int north() {
	return sendMessageAndReceiveAck("T N") ;
    }

    public int south() {
	return sendMessageAndReceiveAck("T S") ;
    }

    public int lw(double width) {
	return sendMessageAndReceiveAck("W " + width) ;
    }

    public int pd() {
	return sendMessageAndReceiveAck("P D") ;
    }

    public int pu() {
	return sendMessageAndReceiveAck("P U") ;
    }

    public int fill() {
	return sendMessageAndReceiveAck("L +") ;
    }

    public int brush() {
	return sendMessageAndReceiveAck("L B") ;
    }

    public int line() {
	return sendMessageAndReceiveAck("L -") ;
    }

    public int point() {
	return sendMessageAndReceiveAck("L P") ;
    }

    public int idle(double t) {
	int rc=0 ;
	try {
	    Thread.sleep((int) (t*1000));
	} catch (InterruptedException e) { rc=-1 ; }
	return rc ;
    }

    public int rf(double range) {
	return sendMessageAndReceiveAck("E " + range) ;
    }

    public int say(String something) {
	return sendMessageAndReceiveAck("S " + convert_str2hex(something)) ;
    }

    public int bcas(String msg) {
	return sendMessageAndReceiveAck("U " + convert_str2hex(msg)) ;
    }

    public int tprintf(String format, Object... varg) {
	String res = String.format(format,varg) ;
	return say(res) ;
    }  

    /* do nothing */
    public void tscanf(String format) {
	return  ;
    }

    public int fire() {
	return sendMessageAndReceiveAck("X")  ;
    }

    public int robot() {
	return sendMessageAndReceiveAck("R") ;
    }

    public int doughnut() {
	return sendMessageAndReceiveAck("D")  ;
    }
    // alias donut doughnut
    public int donut() {
	return doughnut() ;
    }

    public int coin() {
        return sendMessageAndReceiveAck("O X")  ;
    }

    public int borrowcoin() {
        return sendMessageAndReceiveAck("O B")  ;
    }
    public int borrow_coin() {
	return borrow_coin() ;
    }

    public int dropcoin() {
        return sendMessageAndReceiveAck("O S")  ;
    }
    public int drop_coin() {
	return dropcoin() ;
    }

    public int pickcoin() {
        return sendMessageAndReceiveAck("O R")  ;
    }
    public int pick_coin() {
	return pickcoin() ;
    }

    public int col(double r, double g, double b) {
	return sendMessageAndReceiveAck("K " + r + " " + g + " " + b) ;
    }

    public int bgc(double r, double g, double b) {
	return sendMessageAndReceiveAck("H " + r + " " + g + " " + b) ;
    }

    public int card(int c, double x, double y) {
	return sendMessageAndReceiveAck("Z " + c + " " + x + " " + y) ;
    }

    public int clrcard() {
	return sendMessageAndReceiveAck("Z 0") ;
    }

    public int bmode() {
	return sendMessageAndReceiveAck("! +") ;
    }

    public int gmode() {
	return sendMessageAndReceiveAck("! -") ;
    }

    public int nm(String name) {
	return sendMessageAndReceiveAck("N " + convert_str2hex(name)) ;
    }

    public int tm(int team) {
	return sendMessageAndReceiveAck("I " + team) ;
    }

    public int q_id() {
	sendMessageAndReceiveAck("Q M") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public double[] q_pos() {
	sendMessageAndReceiveAck("Q P") ;
	String[] ack = last_message.split(" ",0) ;
	double pos[] = new double[2] ;
	pos[0] = Double.valueOf(ack[0]) ;
	pos[1] = Double.valueOf(ack[1]) ;
	return pos ;
    }

    public double q_dir() {
	sendMessageAndReceiveAck("Q D") ;
	String[] ack = last_message.split(" ",0) ;
	return Double.valueOf(ack[0]) ;
    }
    // alias q_heading q_dir
    public double q_heading() {
	return q_dir() ;
    }

    public boolean q_alive() {
	try {
	    sock_out.println("Q A") ;
	    last_message = sock_in.readLine() ;
	} catch  (IOException e) {
	    return false ;
	}
	if (last_message.length()>0) 
	    return true ;
	else
	    return false  ;
    }

    public int q_score() {
	sendMessageAndReceiveAck("Q S") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public int q_nt() {
	sendMessageAndReceiveAck("Q N") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public int q_coin() {
	sendMessageAndReceiveAck("Q C") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public int q_mycoin() {
	sendMessageAndReceiveAck("Q O") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public double q_radar() {
	sendMessageAndReceiveAck("Q R") ;
	String[] ack = last_message.split(" ",0) ;
	team = Integer.valueOf(ack[1]) ;
	return Double.valueOf(ack[0]) ;
    }

    public double q_radar_n(int nn) {
	sendMessageAndReceiveAck("Q R " + nn) ;
	String[] ack = last_message.split(" ",0) ;
	team = Integer.valueOf(ack[1]) ;
	return Double.valueOf(ack[0]) ;
    }

    public int q_finder() {
	sendMessageAndReceiveAck("Q F") ;
	String[] ack = last_message.split(" ",0) ;
	team = Integer.valueOf(ack[1]) ;
	return Integer.valueOf(ack[0]) ;
    }

    public boolean[] q_sonar() {
	sendMessageAndReceiveAck("Q W") ;
	String[] ack = last_message.split(" ",0) ;
	boolean res[] = new boolean[3] ;
	if (Integer.valueOf(ack[0]) > 0) 
	    res[0] = true ;
	else
	    res[0] = false ;

	if (Integer.valueOf(ack[1]) > 0)
	    res[1] = true ;
	else
	    res[1] = false ;

	if (Integer.valueOf(ack[2]) > 0)
	    res[2] = true ;
	else
	    res[2] = false ;

	return res ;
    }

    public TeamMessage q_bcas() {
	TeamMessage tm = new TeamMessage() ;
	sendMessageAndReceiveAck("Q B") ;
	String[] ack = last_message.split(" ",0) ;
	tm.time = Integer.valueOf(ack[0]) ;
	tm.message = convert_hex2str(ack[1]) ;
	return tm ;
    }

    public int q_time() {
	sendMessageAndReceiveAck("Q T") ;
	String[] ack = last_message.split(" ",0) ;
	return Integer.valueOf(ack[0]) ;
    }

    public int q_stat_card(int c) {
	sendMessageAndReceiveAck("Q Z S " + c) ;
	String[] ack = last_message.split(" ",0) ;
        return Integer.valueOf(ack[0]) ;
    }

    public double[] q_pos_card(int c) {
	sendMessageAndReceiveAck("Q Z P " + c) ;
	String[] ack = last_message.split(" ",0) ;
	double pos[] = new double[2] ;
	pos[0] = Double.valueOf(ack[0]) ;
	pos[1] = Double.valueOf(ack[1]) ;
	return pos ;
    }

}
